"""HTML Fields

The idea is that you can create field values in a standard way and then
you derive a form from the form class to look and behave exactly as you want your form to
behave with your own validation, export, import functions etc. Use a class for each field in the form.

See the doc/examples/code/forms.py example and all will become transparently clear


Developer Ideas
===============

It doesn't matter what the fields do internally but they must expose the following interface:

TIP: It isn't really worth trying to maintain complex data types inside the classes because often the values
come back as strings and your class may expect integers, doing a conversion could lead to an error before the
validation is done and just cause problems. It works out easier just to get the string values from the fields,
maintain them as a list of values and then convert them in the valid() funciton where you can let the user know
if things go wrong. This means clever inheritance is nearly impossible but will proabably save you any hassle.

builderValue() is for other applications to get the final value.


New Ideas
---------

Internally the fields keep their values as string representations. These values are set ant retrieved using the 
set() and get() methods. Fields taking multiple values are kept as an array of strings.

An empty string is treated as a NULL value for single fields. An empty array is treated as NULL for multiple
fields.

The populate() method is used to retrieve information from a dictionary (such as a cgi.FieldStorage() and populate
the field with the appropriate string representation value using the set() method. This way fields don't have to 
have make sure that they submit their string representation when a user clicks submit.

The retrieve() method is used in derived classes to convert from the string representation to a representation
suitatble for the value being represented.

Derived Classes
---------------

Derived from the base classes should be fields with types. For example integer fields, multiple time fields, date
fields etc. These should implement the retrieve() method explained above and the populate() method should also
accept the value output by retrieve().

Attribute Access
----------------

All classes should have private member variables with get() and set() methods. Attribute access should then be
emulated by overriding the __getattr__ () methods. This leads to robust code and a simple interface between the 
variables exposed to the user and the actual variables required.


Typed Fields
============


Public Access
-------------

Only the set() get() and valid() methods should really be used from outside the class. set() and get() can take
the type attribute to over-ride the type of the data in used by these public methods. The type attribute can also
be set in the class constructor.

Formats
-------

Certain fields such as the Time, Date and DateTime fields can display their value in a variety of ways. The format
attribute specified in the constructor or in the html() or frozen() methods defines how these fields display their
values. Eg Time could be an input box requiring strings entered in the format hh:mm. Alternatively it could be two
drop-down boxes, one for hour and one for minute.


Internal Workings
-----------------

Typed fields have a different way of working to the non-typed ones. You specify the values in the correct type
and the field returns that value.
"""

__author__ = "James Gardner <james@xecos.com>"
__date__ = "9th April 2003"
__version__ = "0.1"
__credits__ = """Guido van Rossum, for an excellent programming language."""

import os, cgi, re
import lemon, lemon.date, tags

###########################################################################
#
# Basic Field Classes
#

class Field:
    """Abstract base class from which other fields are derived. Shouldn't be treated as a real field"""
    def __init__(self, name, default, description="", error="", required=False, requiredError='Please enter a value'):

        self.__dict__['_name'] = name
        
        list = []
        if type(default) is type([]) or type(default) is type(()):
            for item in default:
                if item:
                    list.append(str(item))
                else:
                    list.append('')
        else:
            list = str(default)
        self.__dict__['default'] = list
        self.__dict__['requiredError'] = requiredError
        self.__dict__['_value'] = list
        self.__dict__['_description'] = description
        self.__dict__['_error'] = error
        self.__dict__['_type'] = self.__class__.__name__
        self.__dict__['required'] = required
        
    def populate(self, form):
        "Populates the field from a dictionary of cgi variables."

        if form.has_key(self.name()):
            self.set(form[self.name()])
        else:
            self.set('')
        
    def set(self, value):
        "Sets the value of the field manually."
        self._value = value
        
    def valid(self, value):
        if self.required:
            if not value:
                self.setError(self.requiredError)
                return False
        return True
        
    def html(self):
        "Should be overridden."
        raise Exception('The html() function should have been overridden in the derived class.') 
        
    def frozen(self):
        "Should be overridden."
        raise Exception('The frozen() function should have been overridden in the derived class.') 

    def hidden(self):
        "Should be overridden."
        raise Exception('The hidden() function should have been overridden in the derived class.') 
        
    def builderValue(self):
        """Perform any final changes on the finished value then return."""
        return self.get()

    def get(self):
        return self._value
        
    def name(self):
        return self._name
        
    def type(self):  # Unneccessary
        return self._type
        
    def error(self):
        return str(self._error)
        
    def setError(self, value):
        self._error = value
        
    def description(self):
        return self._description
    
    def __getattr__(self, name):
        if name == 'value':
            return self.get()
        else:
            raise Exception("Object does not have a '%s' attribute."%(name))
        
class Input(Field):
    """Text Input Field."""
    def __init__(self, name, default='', description='', error='', size=40, maxlength=None, required=False, requiredError='Please enter a value'):
        """Initialisation code"""
        Field.__init__(self, name, default, description, error, required, requiredError)
        self.__dict__['size'] = size
        self.__dict__['maxlength'] = maxlength

    def populate(self, form):
        if form.has_key(self.name()):
            value = form[self.name()]
            if type(value) is type([]):
                raise KeyError("More than one field named %s has been submitted" % (self.name()))
            else:
                if value:
                    self.set(str(value))
                else:
                    self.set('')
        else:
            self.set('')
        
    def html(self):
        output = ""
        if self.type() is "Password":
            inputType='password'
        elif self.type() is "File":
            inputType='file'
        elif self.type() is "Price":
            output += self.currency + " "
            inputType='text'
        else: 
            inputType='text'
            
        output+='<input type="%s" name="%s"'% (inputType, self.name())
        if type(self.size) is not type(None):
            output+=' size="%s"'% self.size
        if type(self.maxlength) is not type(None):
            output+=' maxlength="%s"'% self.maxlength
        output+=' value="%s">'% self.get()
        return output

    def hidden(self):
        return '<input type="hidden" name="%s" value="%s">'% (self.name(), self.get())

    def frozen(self):
        return self.get()
        
    def set(self, value, dataType=None):
        "Sets the value of the field manually."
        if self.maxlength:
            Field.set(self, value[:self.maxlength])
        else:
            Field.set(self, value)
        
    def valid(self, value):
        if self.required:
            if not value:
                self.setError(self.requiredError)
                return False
        if self.maxlength:
            if len(value)>self.maxlength:
                self.setError("Text must be less then %s characters"%self.maxlength)
                return False
        return True

    def get(self):
        if self.maxlength:
            return self._value[:self.maxlength]
        else:
            return self._value
        
class File(Input):
    """File Input Field."""
    pass
    
class Password(Input):
    """Password Field."""
    pass
    
   
class Hidden(Field):
    """Hidden Field"""
   
   
    def populate(self, form):
        "You cannot populate hidden fields."
        pass
        
    def __init__(self, name, default='', description='', error='', required=False, requiredError='Please enter a value'):
        "Initialisation code"
        Field.__init__(self, name, default, description, error, required, requiredError)
        
    def html(self):
        if self.type() is "CheckBox":
            checked=''
            if self.get() == 'on':
                checked=' checked' 
            output='<input type="checkbox" name="%s"%s>'%(self.name(),checked)
            return output
        else:
            if self.type() is "Button":
                inputType='button'
            elif self.type() is "Submit":
                inputType='submit'
            elif self.type() is "Reset":
                inputType='reset'
            else: 
                inputType='hidden'
                
            output='<input type="%s" name="%s"'% (inputType, self.name())
            output+=' value="%s">'% self.get()
            return output

    def hidden(self):
        return '<input type="hidden" name="%s" value="%s">'% (self.name(), self.get())

    def frozen(self):
        return self.get()
        
  
class CheckBox(Hidden): 
    """CheckBox Field."""
    def frozen(self):
        if self.get() == 'on':
            return "True"
        else:
            return "False"

    def populate(self, form):
        if form.has_key(self.name()):
            value = form[self.name()]
            if type(value) is type([]):
                raise KeyError("More than one field named %s has been submitted" % (self.name()))
            else:
                if (value is not '') or  (overWrite is True):
                    if value:
                        self.set(str(value))
                    else:
                        self.set('')
                else:
                    self.set('')
        else:
            self.set('')
                        
    def valid(self, value):
        if value in ['on','']:
            return True
        self.setError('Please leave the box checked or unchecked.')
        return False
        
class Button(Hidden):
    """Custom Button."""
    pass

class Submit(Hidden):
    """Submit Button."""
    pass

class Reset(Hidden):
    """Reset Button."""
    pass
    
class TextArea(Field):
    """TextArea Field."""
    
    def __init__(self, name, default='', description='', error='', cols=None, rows=None, required=False, requiredError='Please enter some text'):
        """Initialisation code"""
        Field.__init__(self, name, default, description, error, required, requiredError)
        self.__dict__['cols'] = cols
        self.__dict__['rows'] = rows

    def populate(self, form):
        if form.has_key(self.name()):
            value = form[self.name()]
            if type(value) is type([]):
                raise KeyError("More than one field named %s has been submitted" % (self.name()))
            else:
                if value:
                    self.set(str(value).replace('\r\n','\n'))
                else:
                    self.set('')
        else:
            self.set('')
                        
    def html(self):
        output='<textarea name="%s"'% (self.name())
        if type(self.cols) <> type(None):
            output+=' cols="%s"'% self.cols
        if type(self.rows) <> type(None):
            output+=' rows="%s"'% self.rows
        output+='>%s</textarea>'% self.get()
        return output

    def hidden(self):
        return '<input type="hidden" name="%s" value="%s">'% (self.name(), self.get())

    def frozen(self):
        return self.get().replace('\n','<br>')

class Select(Field):
    """Select Field."""
    def __init__(self, name,  options, default='', description='', error='', required=False, requiredError='Please choose an option'):
        """Initialisation code"""
        Field.__init__(self, name, default, description, error, required, requiredError)
        pairs = []
        for pair in options:
            k,v = pair
            pairs.append((str(k), str(v)))
        
        self.__dict__['options'] = pairs
        #self.__dict__['size'] = size
                
    # XXX Maybe override the set method to ensure the value is properly set
    # XXX Rewrite the . access attribute methods.
    
    def populate(self, form):
        if form.has_key(self.name()):
            value = form[self.name()]
            #if type(value) is type([]):
            #    raise TypeError("Multiple values recieved for '%s' when it is set as a single field."%self.name())
            #else:
               
            if value:
                self.set(str(value))
            else:
                self.set('')
        else:
            self.set('')
                            
    def valid(self, value):
        if value == None:
            self.setError("None is not a valid value. This a bug in the software.")
            return False
        elif not value:
            if self.required:
                self.setError(self.requiredError)
                return False
            else:
                return True
        else:
            for option in self.__dict__['options']:
                if str(option[1]) == value:
                    return True
            self.setError("'%s' wasn't one of the options."%value)
            return False

        
        
    def html(self):
    
        output='<select name="%s">\n'%(self.name())

        if len(self.options) > 0:
            for items in self.options:
                selected=''
                k=items[0]
                v=items[1]
                if (type(self.get()) is type(())) or (type(self.get()) is type([])):
                    if v in self.get():
                        selected=" selected"
                else:
                    if v == self.get():
                        selected=" selected"
                output+='  <option value="%s"%s>%s</option>\n'%(v,selected,k)   

        output+='</select>'
        return output

    def hidden(self):
        return '<input type="hidden" name="%s" value="%s">'% (self.name(), self.get())

    def frozen(self):
        if type(self.get()) is type([]):
            results = []
            for option in self.options:
                if str(option[1]) in self.get():
                    results.append(str(option[0]))
            return ", ".join(results)
        else:
            for option in self.options:
                if str(option[1]) == str(self.get()):
                    return str(option[0])
        return self.get() # XXX was return 'XXX Error' but doesn't work with float rounding errors.
            
class Menu(Field):
    """Select Field."""
    def __init__(self, name,  options, default=[], description='', error='', required=False, requiredError='Please choose at least one option', size=None):
        """Initialisation code"""
        Field.__init__(self, name, default, description, error, required, requiredError)
        pairs = []
        for pair in options:
            k,v = pair
            pairs.append((str(k), str(v)))
        
        self.__dict__['options'] = pairs
        self.__dict__['size'] = size
                
    # XXX Maybe override the set method to ensure the value is properly set
    # XXX Rewrite the . access attribute methods.
    
    def populate(self, form):
        if form.has_key(self.name()):
            value = form[self.name()]
            if type(value) is type([]):
                list = []
                for v in value:
                    if v:
                        list.append(str(v))
                    else:
                        list.append('')
                self.set(list)
            else:
                if value:
                    self.set([str(value)])
                else:
                    self.set([])
        else:
            self.set([])
        
    def valid(self, value):
        if len(value) > 0:
            matched = 0
            for val in value:
                for option in self.__dict__['options']:
                    if option[1] == val:
                        matched += 1
            if matched is len(self.get()):
                return True
            else:
                self.setError("One of the options selected is not valid.")
                return False
        elif self.required:
            self.setError(self.requiredError)
            return False
        else:
            return True
        
    def html(self):
        multiple = ' multiple'
        size=' size="%s"'%(self.size)
        output='<select name="%s"%s%s>\n'%(self.name(), size, multiple)
        if self.options > 0:
            for items in self.options:
                selected=''
                k=items[0]
                v=items[1]
                if (type(self.get()) is type(())) or (type(self.get()) is type([])):
                    if v in self.get():
                        selected=" selected"
                else:
                    if v == self.get():
                        selected=" selected"
                output+='  <option value="%s"%s>%s</option>\n'%(v,selected,k)   
        output+='</select>'
        return output

    def hidden(self):
        output = ''
        for value in self.get():
            output += '<input type="hidden" name="%s" value="%s">'% (self.name(), value)
        return output 

    def frozen(self):
        if type(self.get()) is type([]):
            results = []
            for option in self.options:
                if str(option[1]) in self.get():
                    results.append(str(option[0]))
            return ", ".join(results)
        else:
            for option in self.options:
                if str(option[1]) == str(self.get()):
                    return str(option[0])
        return self.get() # XXX was return 'XXX Error' but doesn't work with float rounding errors.
            
            
class RadioGroup(Select):
    """Radio Group Field."""
    
    def __init__(self, name, options, default='', description='', error='', required=False, requiredError='Please choose an option', align='horiz'):
        """Initialisation code"""
        Select.__init__(self, name, options, default, description, error, required, requiredError)
        self.__dict__['align'] = align

    def populate(self, form):
        if form.has_key(self.name()):
            value = form[self.name()]
            if type(value) is type([]):
                raise KeyError("More than one field named %s has been submitted" % (self.name()))
            else:
                if (value is not '') or  (overWrite is True):
                    if value:
                        self.set(str(value))
                    else:
                        self.set('')
                else:
                    self.set('')
                        
    def html(self):
        output=''
        if len(self.options)>0:
            #<select name="%s">\n'%(self.name())
            for option in self.options:
                checked=''
                align=''
                k=option[0]
                v=option[1]
                if str(v)==str(self.get()):
                    checked=" checked"
                if self.align is 'vert':
                    align='<br>'
                output+='  <input type="radio" name="%s" value="%s"%s>%s%s\n'%(self.name(), v, checked, k,align)
        #output+='</select>'
        return output

class CheckBoxGroup(Menu):
    """Check Box Group Field."""
    def __init__(self, name, options, default=None, description='', error='', required=False, requiredError='Please choose at least one option', align='default', cols=4):
        """Initialisation code"""
        Menu.__init__(self, name, options, default, description, error, required, requiredError)
        self.__dict__['align'] = align
        self.__dict__['cols'] = cols


    def populate(self, form):
        if form.has_key(self.name()):
            value = form[self.name()]
            if type(value) is type([]):
                list = []
                for v in value:
                    list.append(str(v))
                self.set(list)
            else:
                if value:
                    self.set([str(value)])
                else:
                    self.set([])
        else:
            self.set([])


    def valid(self, value):
        if (type(value) is type(())) or (type(value) is type([])):
            if len(value) == 0:
                if self.required == True:
                    self.setError(self.requiredError)
                    return False
                else:
                    return True
            else:
                # Check results
                for val in value:
                    strOptions = []
                    for option in self.options:
                        strOptions.append(str(option[1]))
                        
                    if str(val) not in strOptions:
                        self.setError("'%s' is not a valid option."%val)
                        return False
                return True
        raise TypeError('Expected a list object.')
        
    def html(self):
        output = ''
        if len(self.options) > 0:
            if self.align <> 'default':
                for option in self.options:
                    checked=''
                    align=''
                    k=option[0]
                    v=option[1]
                    if v in self.get():
                        checked=" checked"
                    if self.align is 'vert':
                        align='<br>'
                    output+='  <input type="checkbox" name="%s" value="%s"%s>%s%s\n'%(self.name(), v, checked, k,align)
        
            else:
                output += '\n\n    <table border="0" width="100%" cellpadding="0" cellspacing="0">\n    <tr>\n'
                counter = -1
                for option in self.options:
                    counter += 1
                    if ((counter % self.cols) == 0) and (counter <> 0):
                        output += '    </tr>\n    <tr>\n'
                    output += '      <td>'
                    checked=''
                    align=''
                    k=option[0]
                    v=option[1]
                    if v in self.get():
                        checked=" checked"
                    
                    output += '<input type="checkbox" name="%s" value="%s"%s>%s%s'%(self.name(), v, checked, k,align)
                    output += '</td>\n      <td>&nbsp;&nbsp;&nbsp;&nbsp;</td>\n'
                    
                counter += 1
                while (counter % self.cols):
                    counter += 1
                    output += '      <td></td>\n      <td>&nbsp;&nbsp;&nbsp;&nbsp;</td>\n'
                output += '    </tr>\n    </table>\n\n'
        return output


###########################################################################
#
# Combination Field Classes
#

class TypedField(Field):
    """Same as a Field except that no conversion to strings is done. Used as the base class for typed fields."""
    def __init__(self, name, default, description="", error="", required=False, requiredError='Please enter a value'):

        self.__dict__['_name'] = name
        
        list = []
        if type(default) is type([]) or type(default) is type(()):
            for item in default:
                if item:
                    list.append(item)
                else:
                    list.append('')
        else:
            list = default
        self.__dict__['default'] = list
        self.__dict__['requiredError'] = requiredError
        self.__dict__['_value'] = list
        self.__dict__['_description'] = description
        self.__dict__['_error'] = error
        self.__dict__['_type'] = self.__class__.__name__
        self.__dict__['required'] = required
        

        
class String(TypedField):
    "A field for stings of 255 characters or less."
        
    def __init__(self, name, default='', description='', error='', required=False, requiredError='Please enter a value', size=40):
        """Initialisation code"""
        TypedField.__init__(self, name, default, description, error, required, requiredError)
        self.__dict__['size'] = size
        self.__dict__['maxlength'] = 255
        
        if not self.valid(default):
            raise Exception(self.error())

    def populate(self, form):
        if form.has_key(self.name()):
            value = form[self.name()]
            if type(value) is type([]):
                raise KeyError("More than one field named %s has been submitted" % (self.name()))
            else:
                if value:
                    self.set(value)
                else:
                    self.set('')
        else:
            self.set('')
        
    def html(self):
        output='<input type="text" name="%s"'% ( self.name())
        if type(self.size) is not type(None):
            output+=' size="%s"'% self.size
        if type(self.maxlength) is not type(None):
            output+=' maxlength="%s"'% self.maxlength
        output+=' value="%s">'% self.get()
        return output

    def hidden(self):
        return '<input type="hidden" name="%s" value="%s">'% (self.name(), self.get())

    def frozen(self):
        return self.get()
        
    def set(self, value, dataType=None):
        "Sets the value of the field manually."
        if self.maxlength:
            Field.set(self, value[:self.maxlength])
        else:
            Field.set(self, value)

    def valid(self, value):
        if self.required:
            if not value:
                self.setError(self.requiredError)
                return False
        if self.maxlength:
            if len(value)>self.maxlength:
                self.setError("Text must be less then %s characters"%self.maxlength)
                return False
        return True

    def get(self):
        if self.maxlength:
            return self._value[:self.maxlength]
        else:
            return self._value

class Email(String):
    
    def __init__(self, name, default='', description='', error='', required=False, requiredError='Please enter an email address', size=40):
        """Initialisation code"""
        TypedField.__init__(self, name, default, description, error, required, requiredError)
        self.__dict__['size'] = size
        self.__dict__['maxlength'] = None
        
        if default and not self.validEmailAddress(default):
            raise Exception(self.error())

    def validEmailAddress(self, email):
        m = re.search('^.+@.+\..+$',email)
        if not m:
            return False            
        return True
        
    def valid(self, value):
        if self.required and not value:
            self.setError(self.requiredError)
            return False
        if self.maxlength and len(value)>self.maxlength:
            self.setError("Text must be less then %s characters"%self.maxlength)
            return False
        # Validate email address
        if not self.validEmailAddress(value):
            self.setError("Please enter a valid email address.")
            return False            
        return True
class Char(String):
    
    def __init__(self, name, default='', description='', error='', required=False, requiredError='Please enter a value'):
        """Initialisation code"""
        if (type(default)<>type('')) or len(default)>1:
            raise Exception('The default of a Char field should be a string length 1.')
        String.__init__(self, name, default, description, error, required, requiredError, 1 )
        self.__dict__['maxlength'] = 1
        
    def valid(self, value):
        if self.required:
            if not value:
                self.setError(self.requiredError)
                return False
        if self.maxlength:
            if len(value)>self.maxlength:
                self.setError("Characters should be one letter long.")
                return False
        return True
        
class Int(Char):
    
    def __init__(self, name, default='', description='', error='', required=False, requiredError='Please enter a value'):
        """Initialisation code"""
        if (type(default)<>type(1)):
            raise Exception('The default of an Int field should be a valid Python integer. Not a %s.'%type(default))
        String.__init__(self, name, default, description, error, required, requiredError, 11)
        self.__dict__['maxlength'] = 11
        
    def valid(self, value):
        if self.required:
            if not value:
                self.setError(self.requiredError)
                return False
        else:
            try:
                v = int(value)
            except:
                self.setError("Please enter a number from -2147483648 to 2147483647")
                return False
            else:
                if type(v) == type(1):
                    return True
                else:
                    self.setError("Please enter a number from -2147483648 to 2147483647")
                    return False
       
    def get(self):
        try:
            return int(self._value)
        except:
            return self._value
            
class StringSelect(TypedField):
    """StringSelect Field. Differs from a normal Select field in that it only takes the strings as the options.
    The values are assigned by the program."""
    def __init__(self, name,  options, default='', description='', error='', required=False, requiredError='Please enter a value'):
        """Initialisation code"""
        if type(default) <> type(''):
            raise Exception('The default option must be a string.')
        if default not in options:
            raise Exception('The default option must be a one of the options.')
        TypedField.__init__(self, name, default, description, error, required, requiredError)
        for option in options:
            if len(option) > 255:
                raise Exception('String fields should contain strings of 255 characters or less.')
        self.__dict__['options'] = options
    
    def populate(self, form):
        if form.has_key(self.name()):
            value = form[self.name()]
            if value:
                try:
                    self.set(self.options[int(value)])
                except:
                    self.set('')
            else:
                self.set('')
        else:
            self.set('')
                            
    def valid(self, value):
        if value == None:
            self.setError("None is not a valid value")
            return False
        elif not value:
            if self.required:
                self.setError(self.requiredError)
                return False
            else:
                return True
        else:
            if value in self.options:
                return True
            else:
                self.setError("'%s' wasn't one of the options."%value)
                return False
    
    def html(self):
        output='<select name="%s">\n'%(self.name())
        if len(self.options) > 0:
            counter = 0
            for value in self.options:
                selected=''
                if self.get() == value:
                    selected=" selected"
                output+='  <option value="%s"%s>%s</option>\n'%(counter,selected,value)   
                counter += 1
        output+='</select>'
        return output

    def hidden(self):
        return '<input type="hidden" name="%s" value="%s">'% (self.name(), self.get())

    def frozen(self):
        return self.get() 
        
class CharSelect(StringSelect):
        
    def __init__(self, name,  options, default='', description='', error='', required=False, requiredError='Please enter a value'):
        """Initialisation code"""
        if type(default) <> type(''):
            raise Exception('The default option must be a string.')
        if default not in options:
            raise Exception('The default option must be a one of the options.')
        TypedField.__init__(self, name, default, description, error, required, requiredError)
        for option in options:
            if len(option) > 1:
                raise Exception('String fields should contain strings of 1 character or less.')
        self.__dict__['options'] = options
        

        
class StringCheckBoxGroup(StringSelect):
    def __init__(self, name, options, default=[], description='', error='', required=False, requiredError='Please enter a value', align='default', cols=4):
        """Initialisation code"""
        if type(default) <> type([]):
            raise Exception('The default option must be a list of strings.')
        for d in default:
            if d not in options:
                raise Exception('The default option must be a one of the options.')
        TypedField.__init__(self, name, default, description, error, required, requiredError)
        for option in options:
            if len(option) > 255:
                raise Exception('String fields should contain strings of 255 characters or less.')
        self.__dict__['options'] = options
        self.__dict__['align'] = align
        self.__dict__['cols'] = cols


    def populate(self, form):
        if form.has_key(self.name()):
            value = form[self.name()]
            if type(value) is type([]):
                list = []
                for v in value:
                    try:
                        list.append(self.options[int(v)])
                    except:
                        pass
                self.set(list)
            else:
                if value:
                    self.set([self.options[int(value)]])
                else:
                    self.set([])
        else:
            self.set([])


    def valid(self, value):
        if (type(value) is type(())) or (type(value) is type([])):
            if len(value) == 0:
                if self.required == True:
                    self.setError(self.requiredError)
                    return False
                else:
                    return True
            else:
                # Check results
                for val in value:
                    if val not in self.options:
                        self.setError("'%s' is not a valid option."%val)
                        return False
                return True
        raise TypeError('Expected a list object.')
        
    def html(self):
        output = ''
        if len(self.options) > 0:
            if self.align <> 'default':
                counter = 0
                for option in self.options:
                    checked=''
                    align=''
                    k=option
                    v=counter
                    if self.options[v] in self.get():
                        checked=" checked"
                    if self.align is 'vert':
                        align='<br>'
                    output+='  <input type="checkbox" name="%s" value="%s"%s>%s%s\n'%(self.name(), v, checked, k,align)
                    counter+=1
        
            else:
                output += '\n\n    <table border="0" width="100%" cellpadding="0" cellspacing="0">\n    <tr>\n'
                counter = -1
                for option in self.options:
                    counter += 1
                    if ((counter % self.cols) == 0) and (counter <> 0):
                        output += '    </tr>\n    <tr>\n'
                    output += '      <td>'
                    checked=''
                    align=''
                    k=option
                    v=counter
                    if self.options[v] in self.get():
                        checked=" checked"
                    
                    output += '<input type="checkbox" name="%s" value="%s"%s>%s%s'%(self.name(), v, checked, k,align)
                    output += '</td>\n      <td>&nbsp;&nbsp;&nbsp;&nbsp;</td>\n'
                    
                counter += 1
                while (counter % self.cols):
                    counter += 1
                    output += '      <td></td>\n      <td>&nbsp;&nbsp;&nbsp;&nbsp;</td>\n'
                output += '    </tr>\n    </table>\n\n'
        return output


class CharCheckBoxGroup(StringCheckBoxGroup):
    
    def __init__(self, name, options, default=[], description='', error='', required=False, requiredError='Please enter a value', align='default', cols=4):
        """Initialisation code"""
        if type(default) <> type([]):
            raise Exception('The default option must be a list of strings.')
        for d in default:
            if d not in options:
                raise Exception('The default option must be a one of the options.')
        TypedField.__init__(self, name, default, description, error, required, requiredError)
        for option in options:
            if len(option) > 1:
                raise Exception('String fields should contain strings of 1 character or less.')
        self.__dict__['options'] = options
        self.__dict__['align'] = align
        self.__dict__['cols'] = cols
        
class Text(TextArea):
    "A field for stings of any length up to 24bit (ie about 16,000)."
    
    
#~ class Integer(String):
    #~ "A field for integers."
    #~ def populate(self, form):
        #~ if form.has_key(self.name()):
            #~ form[self.name()] = str(form[self.name()])
            
        #~ String.populate(self, form)
            
    #~ def retrieve(self):
        #~ return int(self.get())
        
#~ class float(String):
    #~ "A field for floating point numbers."
    #~ pass
    

class Date(Field):
    """Date Field."""
    def __init__(self, name, default=None, description='', error='', required=False, requiredError='Please enter a value', format='normal'):
        """Initialisation code"""

        if not default:
            default = ''
        
        if type(default) <> type(''): # Assume it is a date object
            try:
                default = default.isoformat()
            except:
                raise Exception(type(default))

        Field.__init__(self, name, default, description, error, required, requiredError)
        self.__dict__['format'] = format
        
        value = self.internal2value(default)
        if not self.valid(value):
            raise Exception(value)

    def populate(self, form):
        if form.has_key(self.name()):
            #if self.format == 'normal':
                #self.set(form[self.name()][6:9]+'-'+form[self.name()][3:4]+'-'+form[self.name()][0:1])
            #else:
            self.set(form[self.name()])
        else:
            self.set('')

    def value2internal(self, value):
        return value[6:10]+'-'+value[3:5]+'-'+value[0:2]
            
    def internal2value(self,value):
        return value[8:10]+'/'+value[5:7]+'/'+value[0:4]
            
    def set(self, value):
        self._value = self.value2internal(value)
            
    def get(self):
        return self.internal2value(self._value)
            
    def valid(self, value):
        if value == '':
            if self.required == True:
                self.setError(self.requiredError)
                return False
            else:
                return True
        elif type(value) <> type(''):
            self.setError('Date values should be strings')
            return False
        return self._validNormal(value)


        
    def _validParts(self, year, month, day):
            if month < 1 or month > 12:
                self.setError('Please choose a valid month.')
                return False
 
            elif day < 1 or day > 31:
                self.setError('Please choose a valid day.')
                return False
            else:
                return True

    def _validISO(self, value):
        if len(value) <> 10:
            self.setError('Values should be in the format yyyy-mm-dd')
            return False
        elif value[4] <> '-' or value[7] <> '-':
            self.setError('Please enter a date in the format yyyy-mm-dd or a blank field')
            return False
        else:
            try:
                year = int(value[0:4])
                month = int(value[5:7])
                day = int(value[8:10])
            except:
                self.setError('Please enter a date in the format yyyy-mm-dd or a blank field')
                return False
            else:
                return self._validParts(year,month,day)
        return True
        
    def _validNormal(self, value):
        if len(value) <> 10:
            self.setError('Values should be in the format dd/mm/yyyy')
            return False
        elif value[2] <> '/' or value[5] <> '/':
            self.setError('Please enter a date in the format dd/mm/yyyy or a blank field')
            return False
        else:
            try:
                year = int(value[6:10])
                month = int(value[3:5])
                day = int(value[0:2])
            except:
                self.setError('Please enter a date in the format dd/mm/yyyy or a blank field')
                return False
            else:
                return self._validParts(year,month,day)
        return True
        
        
    def html(self):
        formatStr='yyyy-mm-dd'
        if self.format == 'normal':
            formatStr='dd/mm/yyyy'
        return Input(self.name(),default=self.get(), maxlength=10, size=10).html() + ' &nbsp; <small>(Format %s)</small>'%(formatStr)
        
    def frozen(self):
        return "%s"%self.get()
        
    def hidden(self):
        return '<input type="hidden" name="%s" value="%s">'%(self.name(), self.get())
        
class Time(Date):
    """Time Field."""
    def __init__(self, name, default=None, description='', error='', required=False, requiredError='Please enter a value'):
        """Initialisation code"""

        if not default:
            default = ''
        
        if type(default) <> type(''): # Assume it is a date object
            default = default.isoformat()


        Field.__init__(self, name, default[:5], description, error, required, requiredError)
        
        if not self.valid(default):
            raise Exception(self.error())
            
    def set(self, value):
        self._value = value
            
    def get(self):
        return self._value

    def populate(self, form):
        if form.has_key(self.name()):
            self.set(form[self.name()])
        else:
            self.set('')
            
    def valid(self, value):
        if value == '':
            if self.required == True:
                self.setError(self.requiredError)
                return False
            else:
                return True
        if type(value) <> type(''):
            self.setError('Time values should be strings.')
            return False
        elif len(value) < 5:# and len(value) <> 5:
            self.setError('Values should be in the format hh:mm') # XXX  or hh:mm.')
            return False
        elif value[2] <> ':':
            self.setError('Please enter a datetime in the format hh:mm or a blank field')
            return False
        else:
            try:
                hour = int(value[0:2])
                if hour < 0 or hour > 23:
                    raise Exception('Invalid hour value')
            except:
                self.setError('Please choose a valid hour')
                return False
                
            try:
                minute = int(value[3:5])
                if minute < 0 or minute > 59:
                    raise Exception('Invalid minutes value')
            except:
                self.setError('Please choose a valid number of minutes')
                return False

            #~ if len(value) == 8:
                #~ try:
                    #~ second = int(value[6:8])
                    #~ if second < 0 or second > 59:
                        #~ raise Exception('Invalid seconds value')
                #~ except:
                    #~ self.setError('Please choose a valid number of seconds')
                    #~ return False
        return True
            
    def html(self):
        return Input(self.name(),default=self.get(), maxlength=10, size=10).html() + ' &nbsp; <small>(Format hh:mm)</small>'
        
           
class DateTime(Date):
    """Time Field."""
    def set(self, value):
        self._value = value
            
    def get(self):
        return self._value
        
    def value2internal(self, value):
        return value
            
    def internal2value(self,value):
        return value
            
            
    def populate(self, form):
        if form.has_key(self.name()):
            self.set(form[self.name()])
        else:
            self.set('')
            
    def valid(self, value):
        if value == '':
            if self.required == True:
                self.setError(self.requiredError)
                return False
            else:
                return True
        elif type(value) <> type(''):
            self.setError('DateTime values should be strings.')
            return False
        elif len(value) <> 19:
            self.setError('Values should be in the format yyyy-mm-dd hh:mm:ss.')
            return False
        elif value[4] <> '-' or value[7] <> '-' or value[10] <> ' ' or value[13] <> ':' or value[16] <> ':':
            self.setError('Please enter a datetime in the format yyyy-mm-dd hh:mm:ss or a blank field')
            return False
        else:
            try:
                year = int(value[0:4])
                month = int(value[5:7])
                day = int(value[8:10])
                hour = int(value[11:13])
                minute = int(value[14:16])
                second = int(value[17:19])
            except:
                self.setError('Please enter a datetime in the format yyyy-mm-dd hh:mm:ss or a blank field. %s'%value)
                return False
            else:
                if hour < 0 or hour > 23:
                    self.setError('Please choose a valid hour.')
                    return False
                elif minute < 0 or minute > 59:
                    self.setError('Please choose a valid number of minutes.')
                    return False
                elif second < 0 or hour > 59:
                    self.setError('Please choose a valid number of seconds.')
                    return False
                elif month < 1 or month > 12:
                    self.setError('Please choose a valid month.')
                    return False
                elif day < 1 or day > 31:
                    self.setError('Please choose a valid day.')
                    return False
                else:
                    return True
        return True


    def html(self):
        return Input(self.name(),default=self.get(), maxlength=19, size=19).html() + ' &nbsp; <small>(Format yyyy-mm-dd hh:mm)</small>'

class Integer(Field):
    pass 