"""Class to handle Persistant HTML forms."""

from fields import *

###########################################################################
#
# Form Classes
#

class Form:
    """Form Class - Designed to be over-ridden to form custom forms made up of html.Field derivatives."""
    
    def __init__(self, name='form', action='', method='post', stickyData={}, enctype='multipart/form-data', populate=None):
        self.__dict__['name'] = name
        self.__dict__['action'] = action
        self.__dict__['method'] = method
        self.__dict__['fields'] = []
        self.__dict__['actions'] = []
        self.__dict__['stickyData'] = stickyData
        self.__dict__['enctype'] = enctype
        
        self.setup()
        if populate:
            self.populate(populate)
        
    def setup(self):
        "Constructor to be overridden in derived classes to construct a form object"
        pass #raise lemon.LemonError('The setup() function should be overridden in the base class.')
        
    def valid(self):
        """Internal function to make sure no wierd values have been set. Populate should take care
        of the standard ones."""
        validates = True
        for field in self.fields:
            #print "<br>Vlaidating Field Name: "+ field.name() + " Field Value: " + str(field.value)
            if not field.valid(field.get()):
                validates = False
        if not self.isValid():
            validates = False
        
        return validates

    def populate(self, form):
        """Sets up the form with the values in the form dictionary. the dictionary should be
        in the form {name1:value1, name2:value2} as the return from the cgi.FieldStorage() 
        function is.
        """
        
        for field in self.fields:
            #if form.has_key(field.name()):
            field.populate(form)
            
    def isValid(self):
        """Function designed to be over-ridden in the derived classes to provide custom validation for
        each individual form."""
        return True
        
    def addField(self, field):
        """Adds the field specified in the method to the fields array of the form."""
        for f in self.fields:
            if f.name() == field.name():
                raise KeyError("The field name '%s' is already in use in this form." % (field.name()))
        self.__dict__['fields'].append(field)
        
    def addAction(self, action):
        """Add a submit button to the actions part of the form with the name 'action'. This is the preferred way
        to add submit buttons rather than specifying them as field objects and calling addField()."""
        for f in self.actions:
            if action == f:
                raise KeyError("The action name '%s' is already in use in this form." % (action))
        self.__dict__['actions'].append(action)
    
    def __getitem__(self, name):
        if type(name) is type(1):
            return self.fields[name]
        else:
            return self.get(name).builderValue()
            
    def get(self, item):
        """If name is a string ,retruns the field object named 'item'.
        Otherwise treats the form as a list of fields in the order they were added and and returns form.fields[item]."""
        stri = ''
        if type(item) is type(1):
            return self.fields[item]
        else:
            for f in self.fields:
                if f.name() == item:
                    return f
                else:
                    stri += ' '+f.name()
                    
        raise ValueError("Field '%s' was not found.%s" %(item,stri))
   
#
# HTML Output Functions
#

    def html(self, text=None, showActions=True):
        if text == None:
            text = self._asFieldsHTML()
        enctype = ''
        if self.enctype is not None:
            enctype = ' enctype="%s"' % self.enctype
        output =  '<form name="%s" class="lemon" action="%s" method="%s"%s>\n'%(self.name, self.action, self.method, enctype)
        output += '<table border="0">\n<tr><td>%s</td></tr>\n'% text
        
        output += lemon.dictAsHiddenFields(self.stickyData)   
        
        if showActions:
            for action in self.actions:
                output += '<tr><td>&nbsp;</td></tr>\n<tr><td><input type="submit" value="%s" name="%s"></td></tr>'%  (action , 'action')
        output += '\n</table>\n</form>'
        return output
        
    def hidden(self):
        return self.html(text=self._asFieldsHidden(), showActions=True)
        
        
    def frozen(self, action=None):
        showActions = True
        if not action:
            showActions = False
        else:
            self.actions = ([action])
        return self.html(self._asFieldsText() + self._asFieldsHidden(), showActions)
        
    #def frozen(self):
    #    return self._asFieldsText()

    def themeDict(self):
        "Return a dictionary suitable for use in a lemon.html.template class."
        fieldsList = []
        for field in self.fields:
            fieldData = []
            fieldData.append(field.name())
            fieldData.append(field.error())
            fieldData.append(field.description())
            fieldData.append(field.builderValue())
            fieldData.append(field.html())
            fieldsList.append(fieldData)
            
        enctype = ''
        if self.enctype is not None:
            enctype = ' enctype="%s"' % self.enctype

        dict = {}
        dict['name'] = self.name()
        dict['action'] = self.action
        dict['method'] = self.method
        dict['enctype'] = enctype
        dict['fields'] = fieldsList
        dict['actions'] = self.actions
        dict['stickyData'] = lemon.dictAsHiddenFields(self.stickyData)
        return dict

    def _asFieldsHTML(self):
        """Outputs the HTML of the form. Desinged to be over-ridden in the derived classes to implement custom
        form output to the web-browser.
        
        Can be accessed either by using the instance attribute .html or using the defualt __str__() method.
        """
        output = '<table border="0" cellpadding="3" cellspacing="0">\n'
        hidden = ''
        #raise TypeError(dir(self.fields[0]))
        for field in self.fields:
            if field.type() <> 'Hidden':
                description  = ''
                errors       = ''
                if field.error():
                    output += '<tr class="error">\n'
                else:
                    output += '<tr>\n'
                output += '  <td valign="top">'
                output += field.description() + ' '
                output += '</td>\n'
                output += '  <td>&nbsp;&nbsp;&nbsp;</td>\n'
                output += '  <td valign="top">' + field.html() + '</td>\n'
                output += '  <td valign="top">'
                output += field.error()
                output += '</td>\n'
                output += '</tr>\n'
            else:
                hidden += field.html()
        output += '</table>\n' + hidden
        return output
        
    def _asFieldsText(self):
        output = '<table border="0" cellpadding="3" cellspacing="0">\n'
        for field in self.fields:
            description  = ''
            errors       = ''
            output += '<tr>\n'
            output += '  <td valign="top">'
            output += field.description() + ' '
            output += '</td>\n'
            output += '  <td>&nbsp;&nbsp;&nbsp;</td>\n'
            output += '  <td valign="top">' + field.frozen() + '</td>\n'
            output += '  <td valign="top">'
            output += field.error()
            output += '</td>\n'
            output += '</tr>\n'
        output += '</table>\n'

        return output
        
    def _asFieldsHidden(self):
        """Outputs the HTML of the form. Desinged to be over-ridden in the derived classes to implement custom
        form output to the web-browser.
        Can be accessed either by using the instance attribute .html or using the defualt __str__() method.
        """
        output = ''
        for field in self.fields:
            output += str(field.hidden()) + '\n'
        return output
        
#
# Value Retrival Functions
#

    def values(self, names=None):
        "Return a tuple containing the values of the form fields in the order they were added."
        valuesList = []
        if names:
            for name in names:
                value = self.get(name).builderValue()
                if value:
                    valuesList.append(value)
                else:
                    valuesList.append('')
        else:
            for field in self.fields:
                value = field.builderValue()
                if value:
                    valuesList.append(value)
                else:
                    valuesList.append('')

        return tuple(valuesList)
        
    def names(self):
        "Return a tuple containing the names of the form fields in the order they were added."
        namesList = []
        for field in self.fields:
            namesList.append(field.name())
        return tuple(namesList)

    def dict(self):
        "Return a dictionary containing the names and valuse of the fields."
        d = {}
        for field in self.fields:
            d[field.name()] = field.builderValue()
        return d

